# -*- coding: utf-8 -*-
#
# QBzr - Qt frontend to Bazaar commands
# Copyright (C) 2009 Gary van der Merwe <garyvdm@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import os
from time import (strftime, localtime)
from PyQt4 import QtCore, QtGui
from bzrlib.workingtree import WorkingTree
from bzrlib.revisiontree import RevisionTree

from bzrlib.plugins.qbzr.lib.cat import QBzrCatWindow
from bzrlib.plugins.qbzr.lib.annotate import AnnotateWindow
from bzrlib.plugins.qbzr.lib.i18n import gettext
from bzrlib.plugins.qbzr.lib.log import LogWindow
from bzrlib.plugins.qbzr.lib.revtreeview import (
    RevisionTreeView,
    RevNoItemDelegate,
    )
from bzrlib.plugins.qbzr.lib.lazycachedrevloader import cached_revisions
from bzrlib.plugins.qbzr.lib.util import (
    get_set_encoding,
    get_summary,
    get_apparent_author_name,
    )
from bzrlib.plugins.qbzr.lib.uifactory import ui_current_widget
from bzrlib.plugins.qbzr.lib.wtlist import ChangeDesc
from bzrlib.plugins.qbzr.lib.subprocess import SimpleSubProcessDialog
from bzrlib.plugins.qbzr.lib.diff import (
    show_diff,
    has_ext_diff,
    ExtDiffMenu,
    InternalWTDiffArgProvider,
    )
from bzrlib.plugins.qbzr.lib.i18n import gettext, N_


class InternalItem():
    __slots__  = ["name", "path", "kind", "file_id"]
    def __init__(self, name, path, kind, file_id):
        self.name = name
        self.path = path
        self.kind = kind
        self.file_id = file_id
    
    revision = property(lambda self:None)

class UnversionedItem(InternalItem):
    def __init__(self, name, path, kind):
        InternalItem.__init__(self, name, path, kind, None)

class ModelItemData():
    __slots__ = ["id", "item", "change", "chekced", "children_ids",
                 "parent_id", "row"]
    
    def __init__(self, item, change):
        self.item = item
        self.change = change
        if change is not None and change.is_ignored() is None:
            self.checked = QtCore.Qt.Checked
        else:
            self.checked = QtCore.Qt.Unchecked
            
        self.children_ids = None
        self.parent_id = None
        self.id = None
        self.row = None

class TreeModel(QtCore.QAbstractItemModel):
    
    HEADER_LABELS = [gettext("File Name"),
                     gettext("Date"),
                     gettext("Rev"),
                     gettext("Message"),
                     gettext("Author"),
                     gettext("Status")]
    NAME, DATE, REVNO, MESSAGE, AUTHOR, STATUS = range(len(HEADER_LABELS))

    REVID = QtCore.Qt.UserRole + 1
    FILEID = QtCore.Qt.UserRole + 2
    PATH = QtCore.Qt.UserRole + 3
    
    def __init__(self, file_icon, dir_icon, symlink_icon, parent=None):
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.file_icon = file_icon
        self.dir_icon = dir_icon
        self.symlink_icon = symlink_icon
        self.tree = None
        self.inventory_data = []
        self.dir_children_ids = {}
        self.parent_ids = []
        self.checkable = False
    
    def set_tree(self, tree, branch=None, 
                 changes_mode=False, want_unversioned=True):
        self.tree = tree
        self.branch = branch
        self.revno_map = None
        self.changes_mode = changes_mode
        
        self.changes = {}
        self.unver_by_parent = {}
        
        if isinstance(self.tree, WorkingTree):
            tree.lock_read()
            try:
                root_id = self.tree.get_root_id()
                for change in self.tree.iter_changes(self.tree.basis_tree(),
                                            want_unversioned=want_unversioned):
                    change = ChangeDesc(change)
                    path = change.path()
                    is_ignored = self.tree.is_ignored(path)
                    change = ChangeDesc(change+(is_ignored,))
                    
                    if change.fileid() is not None and not changes_mode:
                        self.changes[change.fileid()] = change
                    else:
                        if changes_mode:
                            dir_fileid = root_id
                            name = path
                        else:
                            (dir_path, slash, name) = path.rpartition('/')
                            dir_fileid = self.tree.path2id(dir_path)
                        
                        if dir_fileid not in self.unver_by_parent:
                            self.unver_by_parent[dir_fileid] = []
                        
                        if change.is_versioned():
                            item = self.tree.inventory[change.fileid()]
                        else:
                            item = UnversionedItem(name, path, change.kind())
                            
                        self.unver_by_parent[dir_fileid].append(
                                                ModelItemData(item, change))
                
                self.process_inventory(self.working_tree_get_children)
            finally:
                tree.unlock()
        else:
            self.process_inventory(self.revision_tree_get_children)
    
    def revision_tree_get_children(self, item_data):
        for child in item_data.item.children.itervalues():
            yield ModelItemData(child, None)
    
    def working_tree_get_children(self, item_data):
        item = item_data.item
        if isinstance(item, UnversionedItem):
            abspath = self.tree.abspath(item.path)
            
            for name in os.listdir(abspath):
                path = item.path+"/"+name
                (kind,
                 executable,
                 stat_value) = self.tree._comparison_data(None, path)
                child = UnversionedItem(name, path, kind)
                is_ignored = self.tree.is_ignored(path)
                change = ChangeDesc((None,
                                     (None, path),
                                     False,
                                     (False, False),
                                     (None, None),
                                     (None, name),
                                     (None, kind),
                                     (None, executable),
                                     is_ignored))
                yield ModelItemData(child, change)
        
        elif item.children is not None and not self.changes_mode:
            #Because we create copies, we have to get the real item.
            item = self.tree.inventory[item.file_id]
            for child in item.children.itervalues():
                # Create a copy so that we don't have to hold a lock of the wt.
                child = child.copy()
                if child.file_id in self.changes:
                    change = self.changes[child.file_id]
                else:
                    change = None
                yield ModelItemData(child, change)
        if item.file_id in self.unver_by_parent:
            for item_data in self.unver_by_parent[item.file_id]:
                yield item_data
    
    def load_dir(self, dir_id):
        if isinstance(self.tree, WorkingTree):
            self.tree.lock_read()
        try:
            if dir_id>=len(self.inventory_data):
                return
            dir_item = self.inventory_data[dir_id]
            dir_item.children_ids = []
            children = sorted(self.get_children(dir_item),
                              self.inventory_dirs_first_cmp,
                              lambda x: (x.item.name, x.item.kind))
            
            parent_model_index = self._index_from_id(dir_id, 0)
            self.beginInsertRows(parent_model_index, 0, len(children)-1)
            try:
                for child in children:
                    child_id = self.append_item(child, dir_id)
                    dir_item.children_ids.append(child_id)
                    
                    if len(dir_item.children_ids) % 100 == 0:
                        QtCore.QCoreApplication.processEvents()
            finally:
                self.endInsertRows();
        finally:
            if isinstance(self.tree, WorkingTree):
                self.tree.unlock()
    
    def process_inventory(self, get_children):
        self.get_children = get_children
        
        self.emit(QtCore.SIGNAL("layoutAboutToBeChanged()"))
        self.inventory_data = []
        self.dir_children_ids = {}
        self.parent_ids = []
            
        root_item = ModelItemData(
                    self.tree.inventory[self.tree.get_root_id()],
                    None)
        root_id = self.append_item(root_item, None)
        self.load_dir(root_id)
        self.emit(QtCore.SIGNAL("layoutChanged()"))
    
    def append_item(self, item_data, parent_id):
        item_data.id = len(self.inventory_data)
        if parent_id is not None:
            parent_data = self.inventory_data[parent_id]
            if self.is_item_in_select_all(item_data):
                item_data.checked = parent_data.checked
            else:
                item_data.checked = False
            item_data.row = len(parent_data.children_ids)
        else:
            item_data.checked = QtCore.Qt.Checked
            item_data.row = 0
        item_data.parent_id = parent_id
        self.inventory_data.append(item_data)
        return item_data.id
    
    def inventory_dirs_first_cmp(self, x, y):
        (x_name, x_kind) = x
        (y_name, y_kind) = y
        x_a = x_name
        y_a = y_name
        x_is_dir = x_kind =="directory"
        y_is_dir = y_kind =="directory"
        while True:
            x_b, sep, x_a_t = x_a.partition("/")
            y_b, sep, y_a_t = y_a.partition("/")
            if x_a_t == "" and y_a_t == "":
                break
            if (x_is_dir or not x_a_t == "") and not (y_is_dir or not y_a_t == ""):
                return -1
            if (y_is_dir or not y_a_t == "") and not (x_is_dir or not x_a_t == ""):
                return 1
            cmp_r = cmp(x_b, y_b)
            if not cmp_r == 0:
                return cmp_r
            x_a = x_a_t
            y_a = y_a_t
        if x_is_dir and not y_is_dir:
            return -1
        if y_is_dir and not x_is_dir:
            return 1
        return cmp(x_name, y_name)
    
    def set_revno_map(self, revno_map):
        self.revno_map = revno_map
        for item_data in self.inventory_data[1:0]:
            index = self.createIndex (item_data.row, self.REVNO, item_data.id)
            self.emit(QtCore.SIGNAL("dataChanged(QModelIndex, QModelIndex)"),
                      index,index)        
    
    def columnCount(self, parent):
         return len(self.HEADER_LABELS)

    def rowCount(self, parent):
        if parent.internalId()>=len(self.inventory_data):
            return 0
        parent_data = self.inventory_data[parent.internalId()]
        if parent_data.children_ids is None:
            return 0
        return len(parent_data.children_ids)
    
    def canFetchMore(self, parent):
        if parent.internalId()>=len(self.inventory_data):
            return False
        parent_data = self.inventory_data[parent.internalId()]
        return (parent_data.children_ids is None and
                parent_data.item.kind == "directory")
    
    def fetchMore(self, parent):
        self.load_dir(parent.internalId())

    def _index(self, row, column, parent_id):
        if parent_id>=len(self.inventory_data):
            return QtCore.QModelIndex()
        
        parent_data = self.inventory_data[parent_id]
        if parent_data.children_ids is None:
            return QtCore.QModelIndex()
        
        if (row < 0 or
            row >= len(parent_data.children_ids) or
            column < 0 or
            column >= len(self.HEADER_LABELS)):
            return QtCore.QModelIndex()
        item_id = parent_data.children_ids[row]
        return self.createIndex(row, column, item_id)
    
    def _index_from_id(self, item_id, column):
        if item_id >= len(self.inventory_data):
            return QtCore.QModelIndex()
        
        item_data = self.inventory_data[item_id]
        return self.createIndex(item_data.row, column, item_id) 
    
    def index(self, row, column, parent = QtCore.QModelIndex()):
        return self._index(row, column, parent.internalId())
    
    def sibling(self, row, column, index):
        sibling_id = index.internalId()
        if sibling_id == 0:
            return QtCore.QModelIndex()
        parent_id = self.parent_ids[sibling_id]
        return self._index(row, column, parent_id)
    
    def parent(self, child):
        child_id = child.internalId()
        if child_id == 0:
            return QtCore.QModelIndex()
        item_data = self.inventory_data[child_id]
        if item_data.parent_id == 0 :
            return QtCore.QModelIndex()
        
        return self._index_from_id(item_data.parent_id, 0)

    def hasChildren(self, parent):
        if self.tree is None:
            return False
        if parent.internalId() == 0:
            return True
        item_data = self.inventory_data[parent.internalId()]
        return item_data.item.kind == "directory"
    
    is_item_in_select_all = lambda self, item: True
    """Returns wether an item is changed when select all is clicked."""
    
    def setData(self, index, value, role):
        
        
        def set_checked(item_data, checked):
            old_checked = item_data.checked
            item_data.checked = checked
            if not old_checked == checked:
                index = self.createIndex (item_data.row, self.NAME, item_data.id)
                self.emit(QtCore.SIGNAL("dataChanged(QModelIndex, QModelIndex)"),
                          index,index)        
            
        if index.column() == self.NAME and role == QtCore.Qt.CheckStateRole:
            value = value.toInt()[0]
            if index.internalId() >= len(self.inventory_data):
                return False
            
            item_data = self.inventory_data[index.internalId()]
            set_checked(item_data, value)
            
            # Recursively set all children to checked.
            if item_data.children_ids is not None:
                children_ids = list(item_data.children_ids)
                while children_ids:
                    child = self.inventory_data[children_ids.pop(0)]
                    
                    # If unchecking, uncheck everything, but if checking,
                    # only check those in "select_all" get checked.
                    if (self.is_item_in_select_all(child) or
                        value == QtCore.Qt.Unchecked):
                        
                        set_checked(child, value)
                        if child.children_ids is not None:
                            children_ids.extend(child.children_ids)
            
            # Walk up the tree, and update every dir
            parent_data = item_data
            while parent_data.parent_id is not None:
                parent_data = self.inventory_data[parent_data.parent_id]
                has_checked = False
                has_unchecked = False
                for child_id in parent_data.children_ids:
                    child = self.inventory_data[child_id]
                    
                    if child.checked == QtCore.Qt.Checked:
                        has_checked = True
                    elif (child.checked == QtCore.Qt.Unchecked and
                          self.is_item_in_select_all(child)):
                        has_unchecked = True
                    elif child.checked == QtCore.Qt.PartiallyChecked:
                        has_checked = True
                        if self.is_item_in_select_all(child):
                            has_unchecked = True
                    
                    if has_checked and has_unchecked:
                        break
                
                if has_checked and has_unchecked:
                    set_checked(parent_data, QtCore.Qt.PartiallyChecked)
                elif has_checked:
                    set_checked(parent_data, QtCore.Qt.Checked)
                else:
                    set_checked(parent_data, QtCore.Qt.Unchecked)
            
            return True
        
        return False
    
    def data(self, index, role):
        if not index.isValid():
            return QtCore.QVariant()
        
        item_data = self.inventory_data[index.internalId()]
        item = item_data.item
        
        if role == self.FILEID:
            return QtCore.QVariant(item.file_id)
        
        revid = item_data.item.revision
        if role == self.REVID:
            if revid is None:
                return QtCore.QVariant()
            else:
                return QtCore.QVariant(revid)

        column = index.column()
        if column == self.NAME:
            if role == QtCore.Qt.DisplayRole:
                return QtCore.QVariant(item.name)
            if role == QtCore.Qt.DecorationRole:
                if item.kind == "file":
                    return QtCore.QVariant(self.file_icon)
                if item.kind == "directory":
                    return QtCore.QVariant(self.dir_icon)
                if item.kind == "symlink":
                    return QtCore.QVariant(self.symlink_icon)
                return QtCore.QVariant()
            if role ==  QtCore.Qt.CheckStateRole:
                if not self.checkable:
                    return QtCore.QVariant()
                else:
                    return QtCore.QVariant(item_data.checked)
        
        if column == self.STATUS:
            if role == QtCore.Qt.DisplayRole:
                if item_data.change is not None:
                    return QtCore.QVariant(item_data.change.status())
                else:
                    return QtCore.QVariant("")
        
        if column == self.REVNO:
            if role == QtCore.Qt.DisplayRole:
                if self.revno_map is not None and revid in self.revno_map:
                    revno_sequence = self.revno_map[revid]
                    return QtCore.QVariant(
                        ".".join(["%d" % (revno) for revno in revno_sequence]))
                else:
                    return QtCore.QVariant("")

        if role == QtCore.Qt.DisplayRole:
            if revid in cached_revisions:
                rev = cached_revisions[revid]
                
                if column == self.AUTHOR:
                    return QtCore.QVariant(get_apparent_author_name(rev))

                if column == self.MESSAGE:
                    return QtCore.QVariant(get_summary(rev))
        
                if column == self.DATE:
                    return QtCore.QVariant(strftime("%Y-%m-%d %H:%M",   
                                                    localtime(rev.timestamp)))
        
        if role == self.PATH:
            if isinstance(item, UnversionedItem):
                path = item.path
            else:
                if isinstance(self.tree, WorkingTree):
                    self.tree.lock_read()
                try:
                    path = self.tree.id2path(item.file_id)
                finally:
                    if isinstance(self.tree, WorkingTree):
                        self.tree.unlock()
            return QtCore.QVariant(path)
        
        if role == QtCore.Qt.DisplayRole:
            return QtCore.QVariant("")
        return QtCore.QVariant()
    
    def flags(self, index):
        #if not index.isValid():
        #    return QtCore.Qt.ItemIsEnabled

        if self.checkable and index.column() == self.NAME:
            return (QtCore.Qt.ItemIsEnabled |
                    QtCore.Qt.ItemIsSelectable |
                    QtCore.Qt.ItemIsUserCheckable)
        else:
            return (QtCore.Qt.ItemIsEnabled |
                    QtCore.Qt.ItemIsSelectable)

    def headerData(self, section, orientation, role):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return QtCore.QVariant(self.HEADER_LABELS[section])
        return QtCore.QVariant()
    
    def on_revisions_loaded(self, revisions, last_call):
        inventory = self.tree.inventory
        for item_data in self.inventory_data:
            if item_data.id == 0:
                continue
            
            if item_data.item.revision in revisions:
                self.emit(QtCore.SIGNAL("dataChanged(QModelIndex, QModelIndex)"),
                          self.createIndex (item_data.row, self.DATE, item_data.id),
                          self.createIndex (item_data.row, self.AUTHOR, item_data.id))

    def get_repo(self):
        return self.branch.repository
    
    def iter_checked(self):
        return sorted([item_data for item_data in self.inventory_data[1:]
                       if item_data.checked == QtCore.Qt.Checked],
                      self.inventory_dirs_first_cmp,
                      lambda x: (x.change.path(), x.item.kind))

class TreeFilterProxyModel(QtGui.QSortFilterProxyModel):
    source_model = None
    
    filters = [True, True, True, False]
    (UNCHANGED, CHANGED, UNVERSIONED, IGNORED) = range(4)
    
    filter_cache = {}
    
    def setSourceModel (self, source_model):
        self.source_model = source_model
        QtGui.QSortFilterProxyModel.setSourceModel(self, source_model)
    
    def invalidateFilter(self):
        self.filter_cache = {}
        QtGui.QSortFilterProxyModel.invalidateFilter(self)
    
    def setFilter(self, filter, value):
        self.filters[filter] = value
        self.invalidateFilter()
    
    def setFilters(self, filters):
        def iff(b, x, y):
            if b:
                return x
            else:
                return y
        self.filters = [iff(f is not None, f, old_f)
                        for f, old_f in zip(filters, self.filters)]
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row, source_parent):
        if all(self.filters):
            return True
        
        model = self.source_model
        parent_id = source_parent.internalId()
        children_ids = model.inventory_data[parent_id].children_ids
        if len(children_ids)<=source_row:
            return False
        id = children_ids[source_row]
        return self.filter_id_cached(id)
    
    def filter_id_cached(self, id):
        if id in self.filter_cache:
            return self.filter_cache[id]
        else:
            result = self.filter_id(id)
            self.filter_cache[id] = result
            return result
    
    def filter_id(self, id):
        (unchanged, changed, unversioned, ignored) = self.filters

        model = self.source_model
        item_data = model.inventory_data[id]
        
        if item_data.change is None and unchanged: return True
        
        is_versioned = not isinstance(item_data.item, UnversionedItem)
        
        if is_versioned and item_data.change is not None and changed:
            return True
        
        if not is_versioned and unversioned:
            is_ignored = item_data.change.is_ignored()
            if not is_ignored: return True
            if is_ignored and ignored: return True
            if is_ignored and not ignored: return False
        
        if item_data.item.kind == "directory":
            if item_data.children_ids is None:
                model.load_dir(id)
            
            for child_id in item_data.children_ids:
                if self.filter_id_cached(child_id):
                    return True
        
        return False
    
    
    def on_revisions_loaded(self, revisions, last_call):
        self.source_model.on_revisions_loaded(revisions, last_call)
    
    def get_repo(self):
        return self.source_model.get_repo()
    
    def hasChildren(self, parent):
        return self.source_model.hasChildren(self.mapToSource(parent))


class TreeFilterMenu(QtGui.QMenu):
    
    def __init__ (self, parent=None):
        QtGui.QMenu.__init__(self, gettext("&Filter"), parent)
        
        filters = (gettext("Unchanged"),
                   gettext("Changed"),
                   gettext("Unversioned"),
                   gettext("Ignored"))
        
        self.actions = []
        for i, text in enumerate(filters):
            action = QtGui.QAction(text, self)
            action.setData(QtCore.QVariant (i))
            action.setCheckable(True)
            self.addAction(action)
            self.actions.append(action)
        
        self.connect(self, QtCore.SIGNAL("triggered(QAction *)"),
                     self.triggered)
    
    def triggered(self, action):
        filter = action.data().toInt()[0]
        checked = action.isChecked()
        self.emit(QtCore.SIGNAL("triggered(int, bool)"), filter, checked)
    
    def set_filters(self, filters):
        for checked, action in zip(filters, self.actions):
            action.setChecked(checked)


class TreeWidget(RevisionTreeView):

    def __init__(self, *args):
        RevisionTreeView.__init__(self, *args)
        
        self.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
        
        file_icon = self.style().standardIcon(QtGui.QStyle.SP_FileIcon)
        dir_icon = self.style().standardIcon(QtGui.QStyle.SP_DirIcon)
        symlink_icon = self.style().standardIcon(QtGui.QStyle.SP_FileLinkIcon)
        
        self.tree_model = TreeModel(file_icon, dir_icon, symlink_icon)
        self.tree_filter_model = TreeFilterProxyModel()
        self.tree_filter_model.setSourceModel(self.tree_model)
        self.setModel(self.tree_filter_model)
        #self.setModel(self.tree_model)
        
        self.set_header_width_settings()

        self.setItemDelegateForColumn(self.tree_model.REVNO,
                                      RevNoItemDelegate(parent=self))

        self.create_context_menu()
        
        self.connect(self,
                     QtCore.SIGNAL("doubleClicked(QModelIndex)"),
                     self.do_default_action)
        self.default_action = self.show_file_content
        self.tree = None
        self.branch = None
    
    def set_header_width_settings(self):
        header = self.header()
        header.setStretchLastSection(False)
        header.setResizeMode(self.tree_model.NAME, QtGui.QHeaderView.ResizeToContents)
        header.setResizeMode(self.tree_model.DATE, QtGui.QHeaderView.Interactive)
        header.setResizeMode(self.tree_model.REVNO, QtGui.QHeaderView.Interactive)
        header.setResizeMode(self.tree_model.MESSAGE, QtGui.QHeaderView.Stretch)
        header.setResizeMode(self.tree_model.AUTHOR, QtGui.QHeaderView.Interactive)        
        header.setResizeMode(self.tree_model.STATUS, QtGui.QHeaderView.Stretch)        
        fm = self.fontMetrics()
        # XXX Make this dynamic.
        col_margin = 6
        header.resizeSection(self.tree_model.REVNO,
                             fm.width("8888.8.888") + col_margin)
        header.resizeSection(self.tree_model.DATE,
                             fm.width("88-88-8888 88:88") + col_margin)
        header.resizeSection(self.tree_model.AUTHOR,
                             fm.width("Joe I have a Long Name") + col_margin)        
    
    def create_context_menu(self):
        self.context_menu = QtGui.QMenu(self)
        self.action_open_file = self.context_menu.addAction(
                                    gettext("&Open"),
                                    self.open_file)
        self.action_show_file = self.context_menu.addAction(
                                    gettext("&View"),
                                    self.show_file_content)
        self.action_show_annotate = self.context_menu.addAction(
                                    gettext("Show &annotate"), 
                                    self.show_file_annotate)
        self.action_show_log = self.context_menu.addAction(
                                    gettext("Show &log"),
                                    self.show_file_log)
        if has_ext_diff():
            diff_menu = ExtDiffMenu(self)
            self.action_show_diff = self.context_menu.addMenu(diff_menu)
            self.connect(diff_menu,
                         QtCore.SIGNAL("triggered(QString)"),
                         self.show_differences)
        else:
            self.action_show_diff = self.context_menu.addAction(
                                    gettext("Show &differences"),
                                    self.show_differences)
        
        self.context_menu.addSeparator()
        self.action_add = self.context_menu.addAction(
                                    gettext("&Add"),
                                    self.add)
        self.action_revert = self.context_menu.addAction(
                                    gettext("&Revert"),
                                    self.revert)
    
    def set_tree(self, tree, branch=None,
                 changes_mode=False, want_unversioned=True):
        """Causes a tree to be loaded, and displayed in the widget.

        @param changes_mode: If in changes mode, a list of changes, and
                             unversioned items, rather than a tree, is diplayed.
                             e.g., when not in changes mode, one will get:
                             
                             dir1
                                file1              changed
                             file2                 changed
                             
                             but when in changes mode, one will get:
                             
                             dir1/file1             changed
                             file2                  changed
                             
                             When in changes mode, no unchanged items are shown.
        """
        self.tree = tree
        if isinstance(tree, RevisionTree) and branch is None:
            raise AttributeError("A branch must be provided if the tree is a \
                                 RevisionTree")
        self.branch = branch
        self.changes_mode = changes_mode
        self.want_unversioned = want_unversioned
        self.tree_model.set_tree(self.tree, self.branch,
                                 changes_mode, want_unversioned)
        self.tree_filter_model.invalidateFilter()
        
        if str(QtCore.QT_VERSION_STR).startswith("4.4"):
            # 4.4.x have a bug where if you do a layoutChanged when using
            # a QSortFilterProxyModel, it loses all header width settings.
            # So if you are using 4.4, we have to reset the width settings
            # after every time we do a layout changed. The issue is similar to 
            # http://www.qtsoftware.com/developer/task-tracker/index_html?method=entry&id=236755
            self.set_header_width_settings()
        
        header = self.header()
        if isinstance(self.tree, WorkingTree):
            # We currently have to hide the revision columns, because the
            # revision property is not availible from the WorkingTree.inventory.
            # We may be able to get this by looking at the revision tree for
            # the revision of the basis tree.
            header.hideSection(self.tree_model.DATE)
            header.hideSection(self.tree_model.REVNO)
            header.hideSection(self.tree_model.MESSAGE)
            header.hideSection(self.tree_model.AUTHOR)
            header.showSection(self.tree_model.STATUS)
            
            self.context_menu.setDefaultAction(self.action_open_file)
            self.default_action = self.open_file
        else:
            header.showSection(self.tree_model.DATE)
            header.showSection(self.tree_model.REVNO)
            header.showSection(self.tree_model.MESSAGE)
            header.showSection(self.tree_model.AUTHOR)
            header.hideSection(self.tree_model.STATUS)
            
            self.context_menu.setDefaultAction(self.action_show_file)
            self.default_action = self.show_file_content
    
    def refresh(self):
        self.tree_model.set_tree(self.tree, self.branch,
                                 self.changes_mode, self.want_unversioned)
        self.tree_filter_model.invalidateFilter()

    def contextMenuEvent(self, event):
        self.filter_context_menu()
        self.context_menu.popup(event.globalPos())
        event.accept()
    
    def get_selection_indexes(self, indexes=None):
        if indexes == None:
            indexes = self.selectedIndexes()
        rows = {}
        for index in self.selectedIndexes():
            if index.row() not in rows:
                rows[index.internalId()] = index
        return rows.values()
    
    def get_selection_items(self, indexes=None):
        items = []
        if indexes is None:
            indexes = self.get_selection_indexes(indexes) 
        for index in indexes:
            source_index = self.tree_filter_model.mapToSource(index)
            items.append(self.tree_model.inventory_data[
                                                    source_index.internalId()])
        return items

    def filter_context_menu(self):
        is_working_tree = isinstance(self.tree, WorkingTree)
        items = self.get_selection_items()
        versioned = [not isinstance(item.item, UnversionedItem)
                     for item in items]
        changed = [item.change is not None
                   for item in items]
        selection_len = len(items)
        
        single_versioned_file = (selection_len == 1 and versioned[0] and
                                 items[0].item.kind == "file")
        
        self.action_open_file.setEnabled(is_working_tree)
        self.action_open_file.setVisible(is_working_tree)
        self.action_show_file.setEnabled(single_versioned_file)
        self.action_show_annotate.setEnabled(single_versioned_file)
        self.action_show_log.setEnabled(any(versioned))
        self.action_show_diff.setVisible(is_working_tree)
        self.action_show_diff.setEnabled(any(changed))
        
        self.action_add.setVisible(is_working_tree)
        self.action_add.setDisabled(all(versioned))
        self.action_revert.setVisible(is_working_tree)
        self.action_revert.setEnabled(any(changed) and any(versioned))
    
    def do_default_action(self, index=None):
        indexes = self.get_selection_indexes([index])
        if not len(indexes) == 1:
            return
        
        item_data = self.get_selection_items(indexes)[0]
        if item_data.item.kind == "directory":
            # Don't do anything, so that the directory can be expanded.
            return
        
        self.default_action(index)

    @ui_current_widget
    def show_file_content(self, index=None):
        """Launch qcat for one selected file."""
        
        indexes = self.get_selection_indexes([index])
        if not len(indexes) == 1:
            return
        
        item = self.get_selection_items(indexes)[0]
        if isinstance(item.item, UnversionedItem):
            return
        
        index = indexes[0]
        
        path = unicode(index.data(self.tree_model.PATH).toString())

        encoding = get_set_encoding(None, self.branch)
        window = QBzrCatWindow(filename = path,
                               tree = self.tree,
                               parent=self,
                               encoding=encoding)
        window.show()
        self.window().windows.append(window)

    @ui_current_widget
    def open_file(self, index=None):
        """Open the file in the os specified editor."""
        
        if not isinstance(self.tree, WorkingTree):
            raise RuntimeException("Tree must be a working tree to open a file.")
            
        if index is None:
            index = self.currentIndex()
        
        self.tree.lock_read()
        try:
            path = unicode(index.data(self.tree_model.PATH).toString())
            abspath = self.tree.abspath(path)
        finally:
            self.tree.unlock()
        
        url = QtCore.QUrl.fromLocalFile(abspath)
        result = QtGui.QDesktopServices.openUrl(url)

    @ui_current_widget
    def show_file_log(self):
        """Show qlog for one selected file(s)."""
        
        items = self.get_selection_items()
        fileids = [item.item.file_id for item in items
                   if item.item.file_id is not None]
        
        window = LogWindow(None, self.branch, fileids)
        window.show()
        self.window().windows.append(window)
    
    @ui_current_widget
    def show_file_annotate(self):
        """Show qannotate for selected file."""
        index = self.currentIndex()
        file_id = unicode(index.data(self.tree_model.FILEID).toString())
        path = unicode(index.data(self.tree_model.PATH).toString())

        window = AnnotateWindow(self.branch, self.tree, path, file_id)
        window.show()
        self.window().windows.append(window)

    @ui_current_widget
    def show_differences(self, ext_diff=None):
        """Show differences for selected file(s)."""
        
        items = self.get_selection_items()
        if len(items) > 0:
            self.tree.lock_read()
            try:
                # Only paths that have changes.
                paths = [self.tree.id2path(item.item.file_id)
                         for item in items
                         if item.change is not None]
            finally:
                self.tree.unlock()
        else:
            # Show all.
            paths = None
        
        arg_provider = InternalWTDiffArgProvider(
            None, self.tree,
            self.tree.branch, self.tree.branch,
            specific_files=paths)
        
        show_diff(arg_provider, ext_diff=ext_diff,
                  parent_window=self.window())


    @ui_current_widget
    def add(self):
        """Add selected file(s)."""
        
        items = self.get_selection_items()
        
        # Only paths that are not versioned.
        paths = [item.item.path
                 for item in items
                 if isinstance(item.item, UnversionedItem)]
        
        if len(paths) == 0:
            return
        
        args = ["add"]
        args.extend(paths)
        desc = (gettext("Add %s to the tree.") % ", ".join(paths))
        revert_dialog = SimpleSubProcessDialog(gettext("Add"),
                                               desc=desc,
                                               args=args,
                                               dir=self.tree.basedir,
                                               parent=self,
                                               hide_progress=True,)
        res = revert_dialog.exec_()
        if res == QtGui.QDialog.Accepted:
            self.refresh()

    @ui_current_widget
    def revert(self):
        """Revert selected file(s)."""
        
        items = self.get_selection_items()
        
        self.tree.lock_read()
        try:
            # Only paths that have changes.
            paths = [self.tree.id2path(item.item.file_id)
                     for item in items
                     if item.change is not None and
                        not isinstance(item.item, UnversionedItem)]
        finally:
            self.tree.unlock()
        
        if len(paths) == 0:
            return
        
        args = ["revert"]
        args.extend(paths)
        desc = (gettext("Revert %s to latest revision.") % ", ".join(paths))
        revert_dialog = SimpleSubProcessDialog(gettext("Revert"),
                                         desc=desc,
                                         args=args,
                                         dir=self.tree.basedir,
                                         parent=self,
                                         hide_progress=True,)
        res = revert_dialog.exec_()
        if res == QtGui.QDialog.Accepted:
            self.refresh()

class SelectAllCheckBox(QtGui.QCheckBox):
    
    def __init__(self, tree_widget, parent=None):
        QtGui.QCheckBox.__init__(self, gettext("Select / deselect all"), parent)
        
        self.tree_widget = tree_widget
        #self.setTristate(True)
        
        self.connect(tree_widget.tree_model,
                     QtCore.SIGNAL("dataChanged(QModelIndex, QModelIndex)"),
                     self.on_data_changed)
        
        self.connect(self, QtCore.SIGNAL("clicked(bool)"),
                     self.clicked)
    
    def on_data_changed(self, start_index, end_index):
        self.update_state()
    
    def update_state(self):
        model = self.tree_widget.tree_model
        root_index = model._index_from_id(0, model.NAME)
        
        state = model.data(root_index, QtCore.Qt.CheckStateRole)
        self.setCheckState(state.toInt()[0])
    
    def clicked(self, state):
        model = self.tree_widget.tree_model
        root_index = model._index_from_id(0, model.NAME)
        if state:
            state = QtCore.QVariant(QtCore.Qt.Checked)
        else:
            state = QtCore.QVariant(QtCore.Qt.Unchecked)
        
        model.setData(root_index, QtCore.QVariant(state),
                      QtCore.Qt.CheckStateRole)
